#ifndef STK_NOISEWAVETABLE_H
#define STK_NOISEWAVETABLE_H

#include "Generator.h"
#include <stdlib.h>

namespace stk {

/***************************************************/
/*! \class NoiseWavetable
    \brief Periodic repeating noise wavetable

    Wavetable of random noise values genereated by
    C rand() function. Quality of rand() func varies 
    from one OS to another.

    Sample-and-hold wavetable playback behaviour
    Non interpolated, nice zingy frequency aliasing!
*/
/***************************************************/

class NoiseWavetable : public Generator
{
public:

  //! modes set output range 
  static const unsigned int kMODE_BIPOLAR = 0;  //! from -1.0 to +1.0
  static const unsigned int kMODE_UNIPOLAR = 1; //! from  0.0 to +1.0
  
  //! ctor can take mode (see above) and specific seed value.
  /*!
    If the seed value is zero (the default value), the random number generator is
    seeded with the system time.
  */
  NoiseWavetable( unsigned int mode = kMODE_UNIPOLAR, unsigned int seed = 0 );

  //! Seed the random number generator with a specific seed value.
  /*!
    If no seed is provided or the seed value is zero, the random
    number generator is seeded with the current system time.
  */
  void setSeed( unsigned int seed = 0 );

  //! Set the frequency of the periodic noise wavetable playback
  void setFrequency( StkFloat frequency );

  //! Return the last computed output value.
  StkFloat lastOut( void ) const { return lastFrame_[0]; };

  //! Compute and return one output sample.
  StkFloat tick( void );

  //! Fill a channel of the StkFrames object with computed outputs.
  /*!
    The \c channel argument must be less than the number of
    channels in the StkFrames argument (the first channel is specified
    by 0).  However, range checking is only performed if _STK_DEBUG_
    is defined during compilation, in which case an out-of-range value
    will trigger an StkError exception.
  */
  StkFrames& tick( StkFrames& frames, unsigned int channel = 0 );

protected:
  unsigned long sampleHoldPeriod;
  unsigned long sampleHoldRamp;

  static const unsigned int wtSize = 4096;
  StkFloat wavetable[wtSize];
  unsigned int wtReadPos;
};

inline StkFloat NoiseWavetable :: tick( void )
{
  // sample-and-hold wavetable playback behaviour
  // non interpolated, nice zingy frequency aliasing
  if (++sampleHoldRamp > sampleHoldPeriod) {
    sampleHoldRamp = 0; // reset the ramp counter
    wtReadPos++; // increment the wavetable read position
    if (wtReadPos > wtSize) wtReadPos -= wtSize;
  }

  // always return the value at the current read position
  return lastFrame_[0] = wavetable[wtReadPos];
}

inline StkFrames& NoiseWavetable :: tick( StkFrames& frames, unsigned int channel )
{
#if defined(_STK_DEBUG_)
  if ( channel >= frames.channels() ) {
    oStream_ << "Noise::tick(): channel and StkFrames arguments are incompatible!";
    handleError( StkError::FUNCTION_ARGUMENT );
  }
#endif

  StkFloat *samples = &frames[channel];
  unsigned int hop = frames.channels();
  for ( unsigned int i=0; i<frames.frames(); i++, samples += hop )
    *samples = (StkFloat) ( 2.0 * rand() / (RAND_MAX + 1.0) - 1.0 );

  lastFrame_[0] = *(samples-hop);
  return frames;
}

} // stk namespace

#endif //STK_NOISEWAVETABLE_H
