#ifndef VSTXDRUM_H
#define VSTXDRUM_H

/***************************************************/
/*! \class VSTXDrum
    \brief A (very) lofi drum-synth

    This class provides a single voice in the style
    of Vermona DRM, MusicAndMore ADX drum machines.
    Percussive noise with oscillator sweep.

    by phill phelps zenpho@zenpho.co.uk 2015



    History: This previously existed (unpublished)
    as a single-instance-single-voice VST plugin.
    A MaxMSP wrapper assembled a "kit" of drums
    and provided control over the kit via a
    Novation Remote Zero SL MIDI control surface

    See also: VSTXResoFilter, VSTXEnvelope
    VSTXNovationLogic, VSTXNovationDisplay

*/
/***************************************************/

#include "Instrmnt.h"
#include "SineWave.h"
#include "BiQuad.h"
#include "Noise.h"
#include "NoiseWavetable.h"

#include "VSTXResoFilter.h"
#include "VSTXEnvelope.h"

namespace stk {

class VSTXDrum : public Instrmnt
{
 public:
  //! Class constructor.
  VSTXDrum( void );

  //! Class destructor.
  ~VSTXDrum( void );

  //! Start a note with the given frequency and amplitude.
  void noteOn( StkFloat frequency, StkFloat amplitude );

  //! Stop a note with the given amplitude (speed of decay).
  void noteOff( StkFloat amplitude );

  //! how many midi CCs do we support?
  //!   to keep things simple, CCs should be zero-indexed
  //!   with no "unused gaps" in the range of CCs supported
  int getNumControls(void) { return kNumParams; };

  //! Control change specified by \e ccNumber (range 0 to 127)
  //! \e set to the ccValue (range 0.0 to 1.0)
  //!   to keep things simple, CCs should be zero-indexed
  //!   with no "unused gaps" in the range of CCs supported
  void     controlChange ( int ccNumber, StkFloat ccValue );

  //! Inspect the synthesis parameter \e ccNumber (range 0 to 127)
  //! returns value in range 0.0. 1.0 
  //!   to keep things simple, CCs should be zero-indexed
  //!   with no "unused gaps" in the range of CCs supported
  StkFloat controlInspect( int ccNumber );
 
  //! return a string that describes the given /e ccNumber
  char* getParamName( int ccNumber );
  
  //! Compute and return one output sample.
  StkFloat tick( unsigned int channel = 0 );

  //! Fill a channel of the StkFrames object with computed outputs.
  /*!
    The \c channel argument must be less than the number of
    channels in the StkFrames argument (the first channel is specified
    by 0).  However, range checking is only performed if _STK_DEBUG_
    is defined during compilation, in which case an out-of-range value
    will trigger an StkError exception.
  */
  StkFrames& tick( StkFrames& frames, unsigned int channel = 0 );

private:
  // utility functions clip and cube floats
  StkFloat cube(StkFloat input);
  StkFloat clip(StkFloat input);

  // (re)configure envelopes from member var parameters
  void configAllEnvelopes(void);

  enum
  {
    kAllBendRange = 0,
    kBendAtk,
    kBendDcy,

    kNoisBendAmt,
    kNoisFreq,
    kNoisReso,
    kNoisCrushAmt,
    kNoisAtk,
    kNoisDcy,
    kNoisVelo,
    kNoisLevel,

    kOscBendAmt,
    kOscFreq,
    kOscRandAmt,
    kOscRandFreq,
    kOscAtk,
    kOscDcy,
    kOscVelo,
    kOscLevel,

    kVolume,
    kPan,

    kNumParams
  };

  // this is a pointer to an array of string parameter names  
  // we set up the array of strings in ctor
  const char** paramNames;

  StkFloat fAllBendRange;
  StkFloat fBendAtk;
  StkFloat fBendDcy;

  StkFloat fNoisBendAmt;
  StkFloat fNoisFreq;
  StkFloat fNoisReso;
  StkFloat fNoisCrushAmt;
  StkFloat fNoisAtk;
  StkFloat fNoisDcy;
  StkFloat fNoisVelo;
  StkFloat fNoisLevel;

  StkFloat fOscBendAmt;
  StkFloat fOscFreq;
  StkFloat fOscRandAmt;
  StkFloat fOscRandFreq;
  StkFloat fOscAtk;
  StkFloat fOscDcy;
  StkFloat fOscVelo;
  StkFloat fOscLevel;

  // TODO - should these be in the drumkit?
  StkFloat fVolume;
  StkFloat fPan;

  //-----------------------------------------
  // STK generators
  SineWave sineGen;
  NoiseWavetable noiseGen;
  NoiseWavetable jitterGen;

  //-----------------------------------------
  // envelope array
  static const int kNumEnvelopes = 3;
  VSTXEnvelope *pEnvelopes[kNumEnvelopes];

  //-----------------------------------------
  // low pass resonant filter array
  static const int kNumFilters = 1;
  VSTXResoFilter *pResoFilters[kNumFilters];

  //---------------------------------------
  bool areEnvsActive; // amp envs idle?
  float midiVeloAmp;  // amp from midi velo

  //---------------------------------
  // values updated every 400 samples
  long fourHundredCount;

  float AllBendRangeMult;
  float OscFreqParamCubed;
  float OscBendAmtParamCubed;

  //---------------------------------
  // values updated every sample
  float bendEnv;
  float noiseAmpEnv;
  float oscAmpEnv;

  float oscFreq;
  float filtCutoff;

};

// utility cube float
inline StkFloat VSTXDrum :: cube( StkFloat input )
{
	return input * input * input;
}

// utility clip float
inline StkFloat VSTXDrum :: clip( StkFloat input )
{
  if (input >  1.0) input =  1.0;
  if (input < -1.0) input = -1.0;
	return input;
}

// one sample
inline StkFloat VSTXDrum :: tick( unsigned int )
{
  // TODO: arrange an array of oscillator parameters
  // then iterate over both oscs and envelopes

  // update envelope values
  bendEnv     = pEnvelopes[0]->tick(); // * AllBendRangeMult;
  noiseAmpEnv = pEnvelopes[1]->tick();
  oscAmpEnv   = pEnvelopes[2]->tick();

  // apply bend env inversion/scaling
  bendEnv *= AllBendRangeMult;

  // apply bend curvature scaling
  //if ( fAllBendRange > 0.5 || fAllBendRange < -0.5 )
  //{
     //bendEnv = cube(bendEnv);
     //bendEnv = bendEnv * (1.0 - ( 2.0 * AllBendRangeMult ));
     //bendEnv = clip(bendEnv);
  //}
  
  // if !envsActive - this sample should be silent
  float noiseOut = 0.0;
  float sineOut  = 0.0;
  float jitterOut = 0.0;

  // most processing is avoided if amp envelopes became idle earlier
  areEnvsActive = ( pEnvelopes[1]->isActive() || pEnvelopes[2]->isActive() );
  if (areEnvsActive)
  {
      // compute oscillator values
	    noiseOut = noiseGen.tick(); // noise range  0.0 to +1.0
      sineOut  = sineGen.tick();  // sine  range -1.0 to +1.0

      // noise fuzz distortion
      // recursive rectification with normalisation
      // i.e. normalise( abs(noise), -1.0, +1.0) )
      //
      // expensive so we only apply if noise amp envelope is active
      if ( pEnvelopes[1]->isActive() ) {
        int i = 0;
        while (i < (8 - (fNoisCrushAmt * 7)) )
        {
            if (noiseOut < 0.f ) noiseOut = 0.f;
            noiseOut -= 0.5; noiseOut *= 2.0;
            i++;
        }
      }

      if (++fourHundredCount % 400 == 0)
      {
  	// update envelope parameters during notes
	  configAllEnvelopes();


	// --------------------------------------
	// noise S&H frequency
        float noiseRate = (fNoisCrushAmt * fNoisCrushAmt * 2.5) + 0.1;
        noiseGen.setFrequency( noiseRate * Stk::sampleRate() );

        // osc jitter S&H frequency
        jitterGen.setFrequency( cube(fOscRandFreq) * 800 + 5 );

	//---------------------------------------
	// values that are used every sample, but update every 400 samples
	OscBendAmtParamCubed = cube(fOscBendAmt);
	OscFreqParamCubed = cube(fOscFreq);

        // scale the BendRange (B+-) param to have a range -1.0 to +1.0
	AllBendRangeMult = -1.0 + (2.0 * fAllBendRange);

        // scale and clip so that setting control to -0.5 produces -1.0 
        //AllBendRangeMult *= 2;
        //AllBendRangeMult = clip(AllBendRangeMult);

        // configure the noise filter
        filtCutoff  = (8000 * fNoisFreq) + 200; // base frequency from param
        filtCutoff += 4000 * fNoisBendAmt * bendEnv; // offset with envelope
        pResoFilters[0]->filterConfig( filtCutoff, fNoisReso*0.997); // configure filter

  	//oStream_ << "filt: " << filtCutoff << " " << fNoisReso*0.997;
  	//handleError( StkError::STATUS );
      }

      // osc1 jitter random S&H LFO
      jitterOut = jitterGen.tick() * 4000 * cube(fOscRandAmt);

      // compute new osc1 frequency
      oscFreq  = (3900 * OscFreqParamCubed) + 15; // base frequency from param
      oscFreq += jitterOut; // offset with random jitter
      oscFreq += OscBendAmtParamCubed * 4000 * bendEnv; 
      sineGen.setFrequency(oscFreq); // set frequency

      // filter noise
      noiseOut = pResoFilters[0]->filter(noiseOut); //apply filter

      // apply velocity scaling
      noiseOut *= (1.0 - fNoisVelo) * midiVeloAmp * fNoisLevel;
      sineOut  *= (1.0 - fOscVelo) * midiVeloAmp * fOscLevel;

      // apply amplitude envelopes
      noiseOut *= noiseAmpEnv;
      sineOut  *= oscAmpEnv;
  }

  // mix oscillators
  float oscMix;
  oscMix = noiseOut + sineOut;

  // apply pre-clipping gain
  //oscMix = pHiDistPeakFilt->tick( oscMix ); // 4kHz peak
  //oscMix = pLoDistPeakFilt->tick( oscMix ); // 100Hz peak
  //oscMix *= 1 + (5 * fDistort);
  //oscMix = clip(oscMix);

  // clip output
  oscMix = clip(oscMix);

  // output
  lastFrame_[0] = oscMix;

  return lastFrame_[0];
}

inline StkFrames& VSTXDrum :: tick( StkFrames& frames, unsigned int channel )
{
	oStream_ << "VSTXDrum::tick(frames, chan)";
	handleError( StkError::STATUS );

  unsigned int nChannels = lastFrame_.channels();

#if defined(_STK_DEBUG_)
  if ( channel > frames.channels() - nChannels ) {
    oStream_ << "VSTXDrum::tick(): channel and StkFrames arguments are incompatible!";
    handleError( StkError::FUNCTION_ARGUMENT );
  }
#endif

  StkFloat *samples = &frames[channel];
  unsigned int j, hop = frames.channels() - nChannels;
  if ( nChannels == 1 ) {
    for ( unsigned int i=0; i<frames.frames(); i++, samples += hop ) {
      *samples++ = tick(); // mono, so ignore pan
    }
  }
  else {
    for ( unsigned int i=0; i<frames.frames(); i++, samples += hop ) {     
      *samples++ = tick() * (1.0 - fPan);   // left ch inverts pan
      for ( j=1; j<nChannels; j++ )
        *samples++ = lastFrame_[j] * fPan;  // right ch
    }
  }

  return frames;
}

} // stk namespace

#endif
