#ifndef VSTXENVELOPE_HPP
#define VSTXENVELOPE_HPP

/** based on the synthesis tool kit STK
    ADSR envelope class
*/

// require log()
#include <math.h>

const float kMaxTime = 3.0;

class VSTXEnvelope
{
	//! envelope states
	enum {
	    ATTACK,
	    DECAY,
	    SUSTAIN,
	    RELEASE,
	    IDLE
	};

	public:
        /** Default constructor */
        VSTXEnvelope( float sampleRate );

        /** Default destructor */
        ~VSTXEnvelope();

        //! Set target = 1, state = \enum VSTXEnvelope::ATTACK.
        void keyOn( void );

        //! Set target = 0, state = \enum VSTXEnvelope::RELEASE.
        void keyOff( void );

        //! Compute and return one output value
        float tick ( void );

        /** Get last computed output value
         * alias for getCurrent()
         */
        float lastOut ( void ) { return getCurrent(); }

        /** Access state
         * \return The current value of state
         */
        int getState() { return state; }
        /** Set state
         * \param val New value to set
         */
        void setState(int val) { state = val; }

        //! Is the envelope state active or idle?
        bool isIdle()   { return (state == VSTXEnvelope::IDLE); }

        //! Is the envelope state active or idle?
        bool isActive() { return !isIdle(); }

        /** Access target
         * \return The current value of target
         */
        float getTarget() { return target; }
        /** Set target
         * \param val New value to set
         */
        void setTarget(float val) { target = val; }


        /** Access current
         * \return The current value of current
         */
        float getCurrent() { return current; }


        /** Access attackRate
         * \return The current value of attackRate
         */
        float getAttackRate() { return attackRate; }
        /** Set attackRate
         * \param val New value to set
         */
        void setAttackRate(float val) { attackRate = val; }

        /** Access releaseRate
         * \return The current value of releaseRate
         */
        float getDecayRateRate() { return releaseRate; }
        /** Set releaseRate
         * \param val New value to set
         */
        void setDecayRate(float val) { releaseRate = val; }


        /** Access sustainLevel
         * \return The current value of sustainLevel
         */
        float getSustainLevel() { return sustainLevel; }
        /** Set sustainLevel
         * \param val New value to set
         */
        void setSustainLevel(float val) { sustainLevel = val; }


        /** Access releaseRate
         * \return The current value of releaseRate
         */
        float getReleaseRate() { return releaseRate; }
        /** Set releaseRate
         * \param val New value to set
         */
        void setReleaseRate(float val) { releaseRate = val; }

        /** Set envelope ASR values
           no decay
        */
        void setEnvelopeASR( float attackRateVal, float sustainLevelVal, float releaseRateVal )
        {
            attackRate = attackRateVal;
            decayRate = 1.0; // instant decay
            sustainLevel = sustainLevelVal;
            releaseRate = releaseRateVal;
        };

        /** Set envelope ADSR values
        */
        void setEnvelopeADSR( float attackRateVal, float decayRateVal, float sustainLevelVal, float releaseRateVal )
        {
            //cube input values
            attackRateVal = attackRateVal * attackRateVal * attackRateVal;
            decayRateVal = decayRateVal * decayRateVal * decayRateVal;
            sustainLevelVal = sustainLevelVal * sustainLevelVal * sustainLevelVal;
            releaseRateVal = releaseRateVal * releaseRateVal * releaseRateVal;

            // set member variables
            attackRate = attackRateVal;
            decayRate = decayRateVal;
            sustainLevel = sustainLevelVal;
            releaseRate = releaseRateVal;

            // apply scaling
            attackRate = log(1.001 + attackRate);
            attackRate  = attackRate * attackRate * attackRate;

            decayRate  = log(1.001 + decayRate);
            decayRate  = decayRate * decayRate * decayRate;

            releaseRate  = log(1.001 + releaseRate);
            releaseRate  = releaseRate * releaseRate * releaseRate;
        };

    protected:

    private:

        int state; //!< Member variable "state"
        float target; //!< Member variable "target"
        float current; //!< Member variable "current"
        float attackRate; //!< Member variable "attackRate"
        float decayRate; //!< Member variable "decayRate"
        float sustainLevel; //!< Member variable "sustainLevel"
        float releaseRate; //!< Member variable "releaseRate"

        float sampleRate; //! set from samplerate
};

#endif // VSTXENVELOPE_HPP
