// demo.cpp
//
// An example STK program that allows voice playback and control of
// most of the STK instruments.

#define _STK_DEBUG_

#include "Skini.h"
#include "SKINI.msg"
#include "RtAudio.h"

#include "Instrmnt.h"
#include "VSTXDrumKit.h"

#include "Mutex.h"
#include "RtMidi.h"

// Miscellaneous command-line parsing and instrument allocation
// functions are defined in utilites.cpp ... specific to this program.
#include "utilities.h"

#include <signal.h>
#include <iostream>
#include <algorithm>
#include <cmath>
using std::min;

#include "NovationLogic.h"
#include "NovationDisplay.h"

bool done;
static void finish(int ignore){ done = true; }

using namespace stk;

// The TickData structure holds all the class instances and data that
// are shared by the various processing functions.
struct TickData {

  VSTXDrumKit *kit; // a drum kit Stk::Instrmnt

  RtMidiIn *ctlPort;   // midi CC control surface
  RtMidiOut *dispPort; // lcd label cmds for ctl surface
  RtMidiIn *notePort;  // midi Note triggers

  int noteMidiChannel; // listen for notes on this ch (zero-to-15)

  NovationLogic *logic;  // ctl surface behavioural model
  NovationDisplay *disp; // modify lcd screens to label ctls

  int channels; // numAudioChannels
  int counter;  // TODO remove - this was required for SKINI score playback
  bool realtime; // TODO remove - select between score playback / MIDI input

  std::queue<Skini::Message> messageQueue; // midi message queue
  bool haveMessage; // is a midi message waiting in the queue?
  Skini::Message currentMessage; // current message to be processed

  Mutex mutex; // thread-safe locking of the entire TickData struct
  
  // Default constructor.
  TickData()
    : kit(0),
      ctlPort(0), dispPort(0), notePort(0),
      logic(0), disp(0),
      channels(2), counter(0), realtime( false ),
      haveMessage( false ) {}
};

#define DELTA_CONTROL_TICKS 64 // default sample frames between control input checks

// MidiCallback() handles raw RtMidi messages
// and parses them into Skini::Message format
// the notePortMidiCallback() and ctlPortMidiCallback() functions apply
// an appropriate message.srcPort value
Skini::Message midiCallback( 
    double timeStamp, 
    std::vector<unsigned char> *bytes, 
    void *pTickdata )
{  
  Skini::Message message;

  // An empty (or invalid) message is indicated by a type = 0.
  message.type = 0;

  // Parse the MIDI bytes ... discard certain messages
  // we want two-byte messages (e.g. Noteon/off and CCs)
  if ( bytes->size() < 2 ) return message;
  if ( bytes->at(0) > 239 ) return message;

  message.type = bytes->at(0) & 0xF0;
  message.channel = bytes->at(0) & 0x0F;
  
  //std::cout << "MC " << (int)bytes->at(1) << 
  //               " (ch " << (int)message.channel << '\n';

  // realtime messages delta time = 0.0?
  message.time = 0.0; 
  
  message.intValues[0] = bytes->at(1);
  message.floatValues[0] = (StkFloat) message.intValues[0];

  if ( ( message.type != 0xC0 ) && ( message.type != 0xD0 ) ) {
    if ( bytes->size() < 3 ) return message;
    message.intValues[1] = bytes->at(2);
    if ( message.type == 0xE0 ) { // combine pithbend into single "14-bit" value
      message.intValues[0] += message.intValues[1] <<= 7;
      message.floatValues[0] = (StkFloat) message.intValues[0];
      message.intValues[1] = 0;
    }
    else
      message.floatValues[1] = (StkFloat) message.intValues[1];
  }

  return message;
}


// notePortMidiCallback() just queues message returned by midiCallback()
// we do not alter the message.channel value here
//
// this allows processMessage() to know which port the message came from
// i.e. channels 0-15 are notePort chans, and 16-31 are ctlPort chans
void notePortMidiCallback( 
    double timeStamp, 
    std::vector<unsigned char> *bytes, 
    void *pTickdata )
{
  TickData *data = (TickData *) pTickdata;
  Skini::Message message;
  message = midiCallback(timeStamp, bytes, pTickdata);

  data->mutex.lock();
  data->messageQueue.push( message );
  data->mutex.unlock();

}

// ctlPortMidiCallback() sets message channel number to 
// be offset by 16 (i.e. if ctl message arrived on (zero-based) channel 15
// then the queued message will be on (zero-based) channel 31
//
// this allows processMessage() to know which port the message came from
// i.e. channels 0-15 are notePort chans, and 16-31 are ctlPort chans
void ctlPortMidiCallback( 
    double timeStamp, 
    std::vector<unsigned char> *bytes, 
    void *pTickdata )
{
  TickData *data = (TickData *) pTickdata;
  Skini::Message message;
  message = midiCallback(timeStamp, bytes, pTickdata);
  message.channel += 16; // offset ctl message channel value

  data->mutex.lock();
  data->messageQueue.push( message );
  data->mutex.unlock();
}

// The processMessage() function handles notePort MIDI messages
void processMessage( TickData* data )
{
  register StkFloat msgVal1 = data->currentMessage.floatValues[0];
  register StkFloat msgVal2 = data->currentMessage.floatValues[1];
  register StkFloat msgChan = data->currentMessage.channel;

  switch( data->currentMessage.type ) {

  case __SK_Exit_:
    done = true;
    return;

  case __SK_NoteOn_:
      if ( msgVal2 > 0.0 )
      { 
	      if ( msgChan == data->noteMidiChannel ) {
      	  data->kit->noteOn(msgVal1, msgVal2 * ONE_OVER_128);
          // TODO - drum activity indicators
          //int drum = data->kit->getDrumNumForNote(msgVal1);
          //char drumDisp[2];
          //drumDisp[0] = drum + 48; // cunning ascii sprintf %d 
          //drumDisp[1] = '\0';
          //data->disp->lcdTextPos(4, drum, "*");
        } 
        break; // case
      }
      // else a note off, so continue to next case

  case __SK_NoteOff_:
    if ( msgChan == data->noteMidiChannel )
    {
      data->kit->noteOff(msgVal1);
      // TODO - drum activity indicators
      //int drum = data->kit->getDrumNumForNote(msgVal1);
      //data->disp->lcdTextPos(4, drum, " ");  // hide
    }
    break;// case

  case __SK_ControlChange_:
    if ( msgChan >= 16 ) {
      // accept CC messages only on channels >= 16 (zero-based)
      // i.e. channels 0-15 are notePort chans, and 16-31 are ctlPort chans

      // TODO - we might need to create ctlMidiChannel 
      //        and accept CCs from (16 + ctlMidiChannel)
      data->logic->midiCC( (int)msgVal1, (int)msgVal2 );
    }
    break;

  case __SK_AfterTouch_:
    break;

  case __SK_PitchChange_:
    break;

  case __SK_PitchBend_:
    break;

  case __SK_Volume_:
    break;

  case __SK_ProgramChange_:
    break;
  } // end of switch

  data->haveMessage = false; 

  return;
}


// The tick() function handles sample computation and scheduling of message
// processing.  It is called when a new buffer of audio samples is needed.
int tick( void *outputBuffer, void *inputBuffer, unsigned int nBufferFrames,
          double streamTime, RtAudioStreamStatus status, void *dataPointer )
{
  TickData *data = (TickData *) dataPointer;
  register StkFloat sample, *samples = (StkFloat *) outputBuffer;
  int counter, nTicks = (int) nBufferFrames;

  while ( nTicks > 0 && !done ) {

    // if we don't have a message to process
    if ( !data->haveMessage ) {

      // copy message at queuefront to currentMessage and unquque (pop)
      if ( data->messageQueue.size() > 0 ) {
        data->mutex.lock();
        data->currentMessage = data->messageQueue.front();
        data->messageQueue.pop();
        data->mutex.unlock(); 
      }
      else { // else queue is empty so indicate empty invalid msg
        data->currentMessage.type = 0;
        data->haveMessage = false;
      }

      if ( data->currentMessage.type > 0 ) {
        // realtime msgs set counter to zero?
        data->counter = (long) (data->currentMessage.time * Stk::sampleRate());
        data->haveMessage = true;
      }
      else
        data->counter = DELTA_CONTROL_TICKS;
    }

    // realtime msgs set counter to zero?
    counter = min( nTicks, data->counter );
    data->counter -= counter;
    for ( int t=0; t<counter; t++ ) {
      sample = data->kit->tick();  // TODO - stereo

      if ( data->realtime )
        for ( int k=0; k<data->channels; k++ ) *samples++ = sample; // mono
      nTicks--;
    }
    if ( nTicks == 0 ) break;

    // process waiting message
    if ( data->haveMessage ) processMessage(data);

  } // while nticks && !done

  return 0;
}

int main( int argc, char *argv[] )
{
  TickData data;
  RtAudio dac;

  // If you want to change the default sample rate (set in Stk.h), do
  // it before instantiating any objects!  If the sample rate is
  // specified in the command line, it will override this setting.
  Stk::setSampleRate( 44100.0 );

  // Depending on how you compile STK, you may need to explicitly set
  // the path to the rawwave directory.
  Stk::setRawwavePath( "../../rawwaves/" );

  // By default, warning messages are not printed.  If we want to see
  // them, we need to specify that here.
  Stk::showWarnings( true );

  // construct drumkit obj
  data.kit = new VSTXDrumKit;

  data.noteMidiChannel = 9;

  // MIDI control surface port
  // ------ ABORT on fail ------
  data.ctlPort = midiInputPortNamed("ReMOTE ZeRO SL 20:1");
  if ( data.ctlPort )
    data.ctlPort->setCallback( &ctlPortMidiCallback, (void*)&data );
  else {
    data.ctlPort = NULL;
    done = true;
  }

  // MIDI note trigger port
  // ------ ABORT on fail ------
  data.notePort = midiInputPortNamed("ReMOTE ZeRO SL 20:0");
  if ( data.notePort )
    data.notePort->setCallback( &notePortMidiCallback, (void*)&data );
  else {
    data.notePort = NULL;
    done = true;
  }

  // LCD control surface labeling port
  // ------ ABORT on fail ------
  data.dispPort = midiOutputPortNamed("ReMOTE ZeRO SL 20:0");
  if (data.dispPort) {
    // setup the display controller for the midi output port
    data.disp = new NovationDisplay(data.dispPort);

    // setup the logic handler for the control surface
    // we'll call logic.midiCC() later when midi messages arrive
    // - we need to pass pointers to 
    // "data.disp" the display controller
    // "data.kit" the STK::instrmnt drumkit
    // "data.noteMidiChannel" the note trigger midi channel number
    data.logic = new NovationLogic(
	data.disp,
	(VSTXDrumKit*)data.kit,
	&data.noteMidiChannel
	);
  }
  else {
    data.dispPort = NULL;
    done = true;
  }

  // are necessary MIDI ports configured and no errors?
  if ( data.ctlPort &&
       data.notePort &&
       data.dispPort &&
       !done ) {

    // allocate the dac here
    data.realtime = true;
    RtAudioFormat format = ( sizeof(StkFloat) == 8 ) ? 
              RTAUDIO_FLOAT64 : RTAUDIO_FLOAT32;
    RtAudio::StreamParameters parameters;
    parameters.deviceId = dac.getDefaultOutputDevice();
    parameters.nChannels = data.channels; // audio channels

    //unsigned int bufferFrames = RT_BUFFER_SIZE;
    unsigned int bufferFrames = 40;

    try {
      dac.openStream( 
          &parameters, 
          NULL, 
          format, 
          (unsigned int)Stk::sampleRate(), 
          &bufferFrames, 
          &tick, 
          (void *)&data );
    }
    catch ( RtAudioError& error ) {
      error.printMessage();
      goto cleanup;
    }

    // Install an interrupt handler function.
    (void) signal(SIGINT, finish);

    // If realtime output, set our callback function and start the dac.
    try {
      dac.startStream();
    }
    catch ( RtAudioError &error ) {
      error.printMessage();
      goto cleanup;
    }

    // Setup finished.
    while ( !done ) {
      // TODO - abort if the port becomes disconnected
      // isPortOpen doesnt check this - maybe we could try active sensing?
      // if ( !data.ctlPort->isPortOpen() ) done = true;

      // keep the main thread idle
      Stk::sleep( 200 );
    }

    // Shut down the output stream.
    try {
      dac.closeStream();
    }
    catch ( RtAudioError& error ) {
      error.printMessage();
    }
  }

cleanup:
  std::cout << "\nCleaning up...\n";

  if (data.logic) delete data.logic;
  if (data.disp) delete data.disp;
  if (data.dispPort) delete data.dispPort;
  if (data.notePort) delete data.notePort; 
  if (data.ctlPort) delete data.ctlPort;
  if (data.kit) delete data.kit;

  std::cout << "\nGoodbye!\n\n";
  return 0;
}

