/***************************************************/
/*! \class NoiseWavetable
    \brief Periodic repeating noise wavetable

    Wavetable of random noise values genereated by
    C rand() function. Quality of rand() func varies 
    from one OS to another.

    Sample-and-hold wavetable playback behaviour
    Non interpolated, nice zingy frequency aliasing!
*/
/***************************************************/

#include "NoiseWavetable.h"
#include <time.h>

namespace stk {

NoiseWavetable :: NoiseWavetable( unsigned int mode, unsigned int seed )
{
  // Seed the random number generator
  this->setSeed( seed );

  // fill the wavetable with randomly generated noise
  // later we will playback from this table in a cyclic manner
  // creating periodic random noise to shift-register feedback noise
  for(unsigned int i=0; i<wtSize; i++) {
     if ( mode == kMODE_BIPOLAR )
     {
       // range -1.0 to +1.0
       wavetable[i] = (StkFloat) ( 2.0 * rand() / (RAND_MAX + 1.0) - 1.0 );
     }
     else // if ( mode == kMODE_BIPOLAR )
     {
       // range 0.0 to 1.0
       wavetable[i] = (StkFloat)rand()/RAND_MAX;
     }
  }
  
  // set playback position to beginning of the table
  wtReadPos = 0;

  // default playback rate 30Hz repetitions of the table
  sampleHoldRamp = sampleHoldPeriod = 0;
  setFrequency(30);
}

void NoiseWavetable :: setFrequency( StkFloat frequency )
{
  int intFreq = frequency;
  sampleHoldPeriod = Stk::sampleRate() / intFreq;
  if (sampleHoldPeriod < 2) sampleHoldPeriod = 2;
}

void NoiseWavetable :: setSeed( unsigned int seed )
{
  if ( seed == 0 )
    srand( (unsigned int) time( NULL ) );
  else
    srand( seed );
}

} // stk namespace


