/***************************************************/
/*! \class VSTXDrum
    \brief A (very) lofi drum-synth

    This class provides a single voice in the style
    of Vermona DRM, MusicAndMore ADX drum machines.
    Percussive noise with oscillator sweep.

    by phill phelps zenpho@zenpho.co.uk 2015



    History: This previously existed (unpublished)
    as a single-instance-single-voice VST plugin.
    A MaxMSP wrapper assembled a "kit" of drums
    and provided control over the kit via a
    Novation Remote Zero SL MIDI control surface

    See also: VSTXResoFilter, VSTXEnvelope
    VSTXNovationLogic, VSTXNovationDisplay

*/
/***************************************************/

#include "VSTXDrum.h"

namespace stk {

VSTXDrum :: VSTXDrum( void )
{
	fAllBendRange = 1.0;
	fBendAtk = 1.0;
	fBendDcy = 0.3;

	fNoisBendAmt = 1.0;
	fNoisFreq = 0.2;
	fNoisReso = 0.6;
	fNoisCrushAmt = 0.5;
	fNoisAtk = 1.0;
	fNoisDcy = 0.3;
	fNoisVelo = 0.7;
	fNoisLevel = 0.5;

	fOscBendAmt = 0.5;
	fOscFreq = 0.26;
	fOscRandAmt = 0.02;
	fOscRandFreq = 0.5;
	fOscAtk = 1.0;
	fOscDcy = 0.3;
	fOscVelo = 0.7;
	fOscLevel = 0.7;

  fVolume = 1.0;
  fPan = 0.5; // 0.0==left, 0.5==centre, 1.0==right
 
  // kludge to initialise const char* paramNames[]
  static const char* tempNames[kNumParams] = {
    "B+-",
    "BAtk",
    "BDcy",

    "NB~",
    "NFrq",
    "NRso",
    "NCsh",
    "NAtk",
    "NDcy",
    "NVlo",
    "NLvl",

    "OB~",
    "OFrq",
    "OR~",
    "ORSp",
    "OAtk",
    "ODcy",
    "OVlo",
    "OLvl",

    "Vol",
    "Pan"
  };
  paramNames = tempNames;

	float sr = Stk::sampleRate();
		if (sr <= 0) // samplerate may not be set properly in AudioEffectX
	        sr = 44100.0;

	// fill envelope object array
	for (int i=0; i<kNumEnvelopes; i++)
		pEnvelopes[i] = new VSTXEnvelope(sr);
  areEnvsActive = false;

	// fill filter object array
	for (int i=0; i<kNumFilters; i++) {
		pResoFilters[i] = new VSTXResoFilter(sr);
    pResoFilters[i]->filterConfig(fNoisFreq * sr, fNoisReso);
  }

	fourHundredCount = 0;
}

VSTXDrum :: ~VSTXDrum( void )
{
	for (int i=0; i<kNumEnvelopes; i++)
		delete pEnvelopes[i];

	// TODO: different number of reso filters
	for (int i=0; i<kNumFilters; i++)
		delete pResoFilters[i];
}

void VSTXDrum :: configAllEnvelopes()
{
  // update envelope parameters from appropriate member vars
  // f___AtkDcy member vars always have range 0.0 to 1.0
  // local vars here are scaled to our desired range

  // bend envelope
  float bendAtk = (fBendAtk * 0.5) + 0.2;
  float bendDcy = (fBendDcy * 0.5) + 0.2;
	pEnvelopes[0]->setEnvelopeADSR(
					bendAtk,
					bendDcy,
					0.0,
					bendDcy);

	// noise amp envelope
  float noisAtk = (fNoisAtk * 0.5) + 0.2;
  float noisDcy = (fNoisDcy * 0.5) + 0.2;
	pEnvelopes[1]->setEnvelopeADSR(
					noisAtk,
					noisDcy,
					0.0,
					noisDcy);

	// osc amp envelope
  float oscAtk  = (fOscAtk * 0.5) + 0.2;
  float oscDcy  = (fOscDcy * 0.5) + 0.2;
	pEnvelopes[2]->setEnvelopeADSR(
					oscAtk,
					oscDcy,
					0.0,
					oscDcy);

}

void VSTXDrum :: noteOn( StkFloat frequency, StkFloat amplitude )
{
	// amplitude from midi velocity
	midiVeloAmp = amplitude;

	// update envelope parameters before note starts
	configAllEnvelopes();

	pEnvelopes[0]->keyOn();
	pEnvelopes[1]->keyOn();
	pEnvelopes[2]->keyOn();
}

void VSTXDrum :: noteOff( StkFloat amplitude )
{
	//midiVeloAmp = 0; // set to zero to instantly cut off

	pEnvelopes[0]->keyOff();
	pEnvelopes[1]->keyOff();
	pEnvelopes[2]->keyOff();
}

// ////////////////////////////////////////////////////////////////////////

//! Control change specified by \e ccNumber (range 0 to 127)
//! \e set to the ccValue (range 0.0 to 1.0)
//!   to keep things simple, CCs should be zero-indexed
//!   with no "unused gaps" in the range of CCs supported
void VSTXDrum :: controlChange( int number, StkFloat value )
{
#if defined(_STK_DEBUG_)
  if ( Stk::inRange( value, 0.0, 128.0 ) == false ) {
    oStream_ << "VSTXDrum::controlChange: value (" << value << ") is out of range!";
    handleError( StkError::WARNING ); return;
  }
#endif

  //std::cout << "VSTXDrum::controlChange: " << number << " " << value << '\n';   
  
  switch(number)
  {
  case kAllBendRange: fAllBendRange = value; break;
  case kBendAtk:		  fBendAtk = value; break;
  case kBendDcy:		  fBendDcy = value; break;

  case kNoisBendAmt:	fNoisBendAmt = value; break;
  case kNoisFreq:		  fNoisFreq = value; break;
  case kNoisReso:		  fNoisReso = value; break;
  case kNoisCrushAmt: fNoisCrushAmt = value; break;
  case kNoisAtk:		  fNoisAtk = value; break;
  case kNoisDcy:		  fNoisDcy = value; break;
  case kNoisVelo:		  fNoisVelo = value; break;
  case kNoisLevel:		fNoisLevel = value; break;

  case kOscBendAmt:		fOscBendAmt = value; break;
  case kOscFreq:		  fOscFreq = value; break;
  case kOscRandAmt:	  fOscRandAmt = value; break;
  case kOscRandFreq:  fOscRandFreq = value; break;
  case kOscAtk:			  fOscAtk = value; break;
  case kOscDcy:			  fOscDcy = value; break;
  case kOscVelo:		  fOscVelo = value; break;
  case kOscLevel:		  fOscLevel = value; break;

  case kVolume: 		  fVolume = value; break;
  case kPan: 		      fPan = value; break;

#if defined(_STK_DEBUG_)
  default:
    oStream_ << "VSTXDrum::controlChange: undefined control number (" << number << ")!";
    handleError( StkError::WARNING );
    break;
#endif

  }// switch
}

// ////////////////////////////////////////////////////////////////////////

//! Inspect the synthesis parameter \e ccNumber (range 0 to 127)
//! returns value in range 0.0. 1.0 
//!   to keep things simple, CCs should be zero-indexed
//!   with no "unused gaps" in the range of CCs supported
StkFloat VSTXDrum :: controlInspect( int ccNum )
{
  // value overridden by switch-case below
  StkFloat value = 0.0;

  switch(ccNum)
  {
  case kAllBendRange: value = fAllBendRange; break;
  case kBendAtk:      value = fBendAtk; break;
  case kBendDcy:      value = fBendDcy; break;

  case kNoisBendAmt:  value = fNoisBendAmt; break;
  case kNoisFreq:     value = fNoisFreq; break;
  case kNoisReso:     value = fNoisReso; break;
  case kNoisCrushAmt: value = fNoisCrushAmt; break;
  case kNoisAtk:      value = fNoisAtk; break;
  case kNoisDcy:      value = fNoisDcy; break;
  case kNoisVelo:     value = fNoisVelo; break;
  case kNoisLevel:    value = fNoisLevel; break;


  case kOscBendAmt:   value = fOscBendAmt; break;
  case kOscFreq:      value = fOscFreq; break;
  case kOscRandAmt:   value = fOscRandAmt; break;
  case kOscRandFreq:  value = fOscRandFreq; break;
  case kOscAtk:       value = fOscAtk; break;
  case kOscDcy:       value = fOscDcy; break;
  case kOscVelo:      value = fOscVelo; break;
  case kOscLevel:     value = fOscLevel; break;

  case kVolume:       value = fVolume; break;
  case kPan:          value = fPan; break;


#if defined(_STK_DEBUG_)
  default:
    oStream_ << "VSTXDrum::controlChange: undefined control number (" << ccNum << ")!";
    handleError( StkError::WARNING );
    break;
#endif
  }// switch

  return value;
}

//! return a string that describes the given /e ccNum
char* VSTXDrum::getParamName( int ccNum )
{
  if ( ccNum < 0 ) ccNum = 0;
  if ( ccNum > kNumParams-1 ) ccNum = kNumParams-1;
  return (char*) paramNames[ccNum];
}

} // stk namespace
